with Claw.Sockets,
     Text_Line_Sockets;
package SMTP is    -- Send Mail, RFC 821

  type Socket_Type is new Text_Line_Sockets.Socket_Type(200) with private;

  Response_Error: exception; -- raised by any unexpected response
                             -- use Get_Response_Code to get the code
                             -- Socket will automatically be closed.

  procedure Open(Socket : in out Socket_Type;
                 Host   : in     String;  -- eg "www.yahoo.com", "11.22.33.44"
                 Port   : in     Claw.Sockets.Port_Type := Claw.Sockets.mail_port;
                 Patience:in     Duration := 30.0);
    -- Raises Claw.Not_Found_Error if it can't find Host.

  procedure Open(Socket : in out Socket_Type;
                 IP     : in     Claw.Sockets.Network_Address_Type;
                 Host_Name : in  String; -- for "Host: www.a.b" in Get/etc
                 Port   : in     Claw.Sockets.Port_Type := Claw.Sockets.mail_port;
                 Patience:in     Duration := 30.0);

  subtype Response_Code_Type is String(1 .. 3);

  procedure Get_Response_Code(Socket : in out Socket_Type;
                              Response_Code : out Response_Code_Type);
  -- This call returns the most recent response code, and resets to blank.

  procedure Get_Error_Code(Socket : in out Socket_Type;
                           Error_Code :out Claw.Sockets.Error_Codes);
  -- Note: Claw.Sockets.Image(Error_Code) returns a String.
  -- This call resets the saved error code to zero.

  procedure Send_Message_Info(Socket : in out Socket_Type;
                              Sender : in     String;
                              To     : in     String;
                              CC, BCC, From, Reply_To : in String := "";
                              Subject : in String;
                              Header_Info_Continuing : Boolean := False);
    -- If From is null, Sender will be used.
    -- If Reply_To is null, From or Sender will be implied.
    -- If Header_Info_Continuing is True, then you may continue
    -- (via Put_Content_Line calls) to send header information lines,
    -- per RFC 822, finally ending with Put_Content_Line(Socket, Line=>"");

  procedure Put_Content_Line(Socket : in out Socket_Type;
                             Line : in String);

  procedure Finish_Message(Socket : in out Socket_Type);

  procedure Close(Socket : in out Socket_Type);
    -- Closes the Socket.  Open is the sole legal next operation.

private

  type States is (Closed, Ready_To_Send_Header, Header_Sent);

  type Socket_Type is new Text_Line_Sockets.Socket_Type(200)
--  (Max_Line_Length => 120) -- allow moderate sized response lines
  with record
    Mail_State : States := Closed;
    Patience : Duration;
    Error_Code : Claw.Sockets.Error_Codes :=0;
    Response_Code : Response_Code_Type := "   ";
  end record;

  procedure Finalize(Socket : in out Socket_Type);

end SMTP;
