with Ada.Characters.Handling,
     Ada.Characters.Latin_1,
     Interfaces,
     Ada.Streams.Stream_IO,
     Ada.Unchecked_Conversion;
package body WORM_Str is

  function Count(Store : Store_Type) return String_Counts is
  -- return number of stored strings.
  begin return Store.Next_String-1;end Count;

  function Character_Count(Store : Store_Type) return String_Character_Counts is
  -- return number of characters stored.
  begin return Store.Last_List(Store.Next_String-1);end Character_Count;

  procedure Do_Add_String(Store : in out Store_Type;
                          S : in String;
                          Result : out String_Indices) is
  -- Add to store. Return new index.
  -- Raise Overflow if too many or too long strings.
    Start : Data_Indices;
  begin
    if Store.Next_String = String_Indices'first then  -- initialize
      Store.Last_List(String_Indices'first-1) := Store.Data'first;
      Store.Data(Store.Data'first) := ' ';
    end if;
    if Store.Next_String > Store.Last_List'last then
      raise Overflow;
    end if;
    if Store.Last_List(Store.Next_String-1)+S'length > Store.Data'last then
      raise Overflow;
    end if;
    Start := Store.Last_List(Store.Next_String-1)+1;
    Store.Last_List(Store.Next_String) := Start+S'length-1;
    Store.Data(Start .. Store.Last_List(Store.Next_String)) := Data_Type(S);
    Store.Next_String := Store.Next_String+1;
    Result := Store.Next_String-1;
  end Do_Add_String;

  protected Safe_Writer is
    procedure Add_String(Store : in out Store_Type;
                         S : in String; Result : out String_Indices);
  end Safe_Writer;

  protected body Safe_Writer is
    procedure Add_String(Store : in out Store_Type;
                         S : in String; Result : out String_Indices) is
    begin Do_Add_String(Store, S, Result);end Add_String;
  end Safe_Writer;

  procedure Add(Store : in out Store_Type;
                S : in String;
                Index : out String_Indices) is
  begin
    if Store.Task_Safe then
      Safe_Writer.Add_String(Store, S, Index);
    else
      Do_Add_String(Store, S, Index);
    end if;
  end Add;

  function Get(Store : Store_Type;
               Index : in String_Indices) return String is
  -- raise Nonesuch if no such string.
    Start : Data_Indices;
  begin
    if Index >= Store.Next_String then raise Nonesuch;end if;
    Start := Store.Last_List(Index-1)+1;
    declare
      Result : String(1 .. Integer(Store.Last_List(Index)-Start+1));
      -- get indices right
    begin
      for i in Result'range loop
        Result(i) := Store.Data(Data_Indices(i)+Start-1);
      end loop;
      return Result;
    end;
  end Get;

  function Get_Length(Store : Store_Type;
                      Index : in String_Indices) return Natural is
  -- raise Nonesuch if no such string.
    Start : Data_Indices;
  begin
    if Index >= Store.Next_String then raise Nonesuch;end if;
    Start := Store.Last_List(Index-1)+1;
    return Integer(Store.Last_List(Index)-Start+1);
  end Get_Length;

--procedure dump(ll : in ada.streams.stream_element_array) is
--begin
--  for i in ll'range loop
--    log.note(ada.streams.stream_element_offset'image(i)
--             & ada.streams.stream_element'image(ll(i)));
--  end loop;
--end dump;

  subtype Four_Bytes is Ada.Streams.Stream_Element_Array(1 .. 4);

  subtype Buffer_Type is Ada.Streams.Stream_Element_Array(1 .. 16384);
  -- Must be multiple of Four_Bytes

  function Convert is new Ada.Unchecked_Conversion(Interfaces.Integer_32,
                                                   Four_Bytes);
  function Convert is new Ada.Unchecked_Conversion(Four_Bytes,
                                                   Interfaces.Integer_32);

  function External_String_Counts(X : String_Counts) return Four_Bytes is
  begin
    return Convert(Interfaces.Integer_32(X));
  end External_String_Counts;

  function External_String_Character_Counts(X : String_Character_Counts)
    return Four_Bytes is
  begin
    return Convert(Interfaces.Integer_32(X));
  end External_String_Character_Counts;

  function Internal_String_Counts(X : Four_Bytes) return String_Counts is
  begin
    return String_Counts(Interfaces.Integer_32'(Convert(X)));
  end Internal_String_Counts;

  function Internal_String_Character_Counts(X : Four_Bytes)
    return String_Character_Counts is
  begin
    return String_Character_Counts(Interfaces.Integer_32'(Convert(X)));
  end Internal_String_Character_Counts;

  procedure Last_List_Writer(Stream : access Ada.Streams.Root_Stream_Type'class;
                             Info : in Last_List_Type) is
    use Ada.Streams;
    di : Stream_Element_Offset;
    Four : Four_Bytes;
    Buffer : Buffer_Type;
  begin
    di := 1;
    for i in Info'range loop
      if di > Buffer'last then
        Write(Stream.all, Buffer);
        di := 1;
      end if;
      Four := External_String_Character_Counts(Info(i));
      Buffer(di .. di+3) := Four;
      di := di+4;
    end loop;
    if di > 1 then
      Write(Stream.all, Buffer(1 .. di-1));
    end if;
  end Last_List_Writer;

  procedure Last_List_Reader(Stream : access Ada.Streams.Root_Stream_Type'class;
                        Info : out Last_List_Type) is
    use Ada.Streams;
    Last,Last_To_Get : Stream_Element_Offset;
    si : Stream_Element_Offset;
    Four : Four_Bytes;
    Buffer : Buffer_Type;
  begin
    si := Buffer'last+1;
    for i in Info'range loop
      if si > Buffer'last then
        Last_To_Get := Stream_Element_Offset'min(Buffer'last,
                         4*Stream_Element_Offset(Info'last-i+1));
        Read(Stream.all, Buffer(1 .. Last_To_Get), Last);
        if Last /= Last_To_Get then
          raise Ada.Streams.Stream_IO.Data_Error;
        end if;
        si := 1;
      end if;
      Four := Buffer(si .. si+3);
      Info(i) := Internal_String_Character_Counts(Four);
      si := si+4;
    end loop;
  end Last_List_Reader;

  procedure Data_Writer(Stream : access Ada.Streams.Root_Stream_Type'class;
                        Info : in Data_Type) is
    use Ada.Streams;
    di : Stream_Element_Offset;
    Buffer : Buffer_Type;
  begin
    di := 1;
    for i in Info'range loop
      if di > Buffer'last then
        Write(Stream.all, Buffer);
        di := 1;
      end if;
      Buffer(di) := Stream_Element(Character'pos(Info(i)));
      di := di+1;
    end loop;
    if di > 1 then
      Write(Stream.all, Buffer(1 .. di-1));
    end if;
  end Data_Writer;

  procedure Data_Reader(Stream : access Ada.Streams.Root_Stream_Type'class;
                        Info : out Data_Type) is
    use Ada.Streams;
    Last,Last_To_Get : Stream_Element_Offset;
    si : Stream_Element_Offset;
    Buffer : Buffer_Type;
  begin
    si := Buffer'last+1;
    for i in Info'range loop
      if si > Buffer'last then
        Last_To_Get := Stream_Element_Offset'min(Buffer'last,
                         Stream_Element_Offset
                           (Info'last-i+1));
        Read(Stream.all, Buffer(1 .. Last_To_Get), Last);
        if Last /= Last_To_Get then
          raise Ada.Streams.Stream_IO.Data_Error;
        end if;
        si := 1;
      end if;
      Info(i) := Character'val(Buffer(si));
      si := si+1;
    end loop;
  end Data_Reader;

  procedure Store_Type_Write(Stream : access Ada.Streams.Root_Stream_Type'class;
                             Item : in Store_Type) is
    use Ada.Streams;
    Last_Data_Index : String_Character_Counts;
  begin
    Write(Stream.all, External_String_Counts(Item.Next_String-1));
    if Item.Next_String = String_Indices'first then return;end if;
    Last_Data_Index := Item.Last_List(Item.Next_String-1);
    Write(Stream.all, External_String_Character_Counts(Last_Data_Index));
    Last_List_Writer(Stream, Item.Last_List(0 .. Item.Next_String-1));
    Data_Writer(Stream, Item.Data(1 .. Last_Data_Index));
  end Store_Type_Write;

  procedure Store_Type_Read(Stream : access Ada.Streams.Root_Stream_Type'class;
                            Item   : out    Store_Type) is
    use Ada.Streams;
    Next_String : String_Indices;
    Last : Stream_Element_Offset;
    Last_Data_Index : String_Character_Counts;
    Four : Four_Bytes;
  begin
    Read(Stream.all, Four, Last);
    if Last /= Four'last then
      raise Ada.Streams.Stream_IO.Data_Error;
    end if;
    Next_String := Internal_String_Counts(Four)+1;
    if Next_String-1 > Item.Last_List'last then
      raise Constraint_Error;
    end if;
    if Next_String = String_Indices'first then -- empty string store
      Item.Next_String := String_Indices'first;
      return;
    end if;
    Read(Stream.all, Four, Last);
    if Last /= Four'last then
      raise Ada.Streams.Stream_IO.Data_Error;
    end if;
    if Item.Data'last < Internal_String_Character_Counts(Four) then
      raise Constraint_Error;
    end if;
    -- appears to fit.
    Item.Next_String := Next_String;
    Last_Data_Index := Internal_String_Character_Counts(Four);
    Last_List_Reader(Stream, Item.Last_List(0 .. Item.Next_String-1));
    Data_Reader(Stream, Item.Data(1 .. Last_Data_Index));
  end Store_Type_Read;

  procedure Comparison_Map_Type_Write(Stream : access Ada.Streams.Root_Stream_Type'class;
                                      Item   : in     Comparison_Map_Type) is
  -- for Comparison_Map_Type'Write use Comparison_Map_Type_Write
    use Ada.Streams;
    Buffer : Stream_Element_Array(1 .. Stream_Element_Offset(Item'length));
    di : Stream_Element_Offset := 1;
  begin
    for i in Item'range loop
      Buffer(di) := Stream_Element(Character'pos(Item(i)));
      di := di+1;
    end loop;
    Write(Stream.all, Buffer);
  end Comparison_Map_Type_Write;

  procedure Comparison_Map_Type_Read(Stream : access Ada.Streams.Root_Stream_Type'class;
                                     Item   : out    Comparison_Map_Type) is
  -- for Comparison_Map_Type'Read use Comparison_Map_Type_Read
    use Ada.Streams;
    Buffer : Stream_Element_Array(1 .. Stream_Element_Offset(Item'length));
    si : Stream_Element_Offset := 1;
    Last : Stream_Element_Offset;
  begin
    Read(Stream.all, Buffer, Last);
    if Last /= Buffer'last then
      raise Ada.Streams.Stream_IO.Data_Error;
    end if;
    for i in Item'range loop
      Item(i) := Character'val(Buffer(si));
      si := si+1;
    end loop;
  end Comparison_Map_Type_Read;

  function Case_Sensitive_Compare return Comparison_Map_Type is
    Result : Comparison_Map_Type;
  begin
    for c in Result'range loop
      Result(c) := c;
    end loop;
    return Result;
  end Case_Sensitive_Compare;

  function Case_Insensitive_Compare return Comparison_Map_Type is
    use Ada.Characters.Handling;
    Result : Comparison_Map_Type;
  begin
    for c in Result'range loop
      Result(c) := To_Upper(c);
    end loop;
    return Result;
  end Case_Insensitive_Compare;

  function aAbBcC_Compare return Comparison_Map_Type is
    use Ada.Characters.Handling;
    use Ada.Characters.Latin_1;
    Result,
    Inverse : Comparison_Map_Type;
    di : Character;
  begin
    for c in Result'range loop
      Result(c) := c;
      Inverse(c) := c;
    end loop;
    -- replace 'A' .. 'Z', Left_Square_Bracket .. Grave, 'a' .. 'z'
    -- with "aAbB..zZ", Left_Square_Bracket .. Grave
    Inverse('u' .. 'z') := Inverse(Left_Square_Bracket .. Grave);
    di := 'A';
    for i in Character range 'A' .. 'Z' loop
      Inverse(di) := To_Lower(i);
      di := Character'succ(di);
      Inverse(di) := i;
      di := Character'succ(di);
    end loop;
    for c in character loop
      Result(Inverse(c)) := c;
    end loop;
    return Result;
  end aAbBcC_Compare;

  function Compare(Store : Store_Type;
                   Comparison_Map : Comparison_Map_Type;
                   Index : String_Indices; S : String)
  return Comparisons is
    use Ada.Characters.Handling;
    SI : Data_Indices;
    Len : Natural;
    j : Integer := S'first;
    UD : Character;
    US:  Character;
  begin
    if Index >= Store.Next_String then raise Nonesuch;end if;
    SI := Store.Last_List(Index-1)+1;
    Len := Natural(Store.Last_List(Index) - Store.Last_List(Index-1));
    if Len = S'length
    and then Data_Type(S) = Store.Data(SI .. Store.Last_List(Index)) then
      return EQ;
    end if;
    if S'length = 0 then return GT;end if;
    if Len = 0 then return LT;end if;
    -- not EQ and neither is zero length
    if Comparison_Map(Store.Data(SI)) > Comparison_Map(S(S'first)) then return GT;
    elsif Comparison_Map(Store.Data(SI)) < Comparison_Map(S(S'first)) then return LT;
    else
      for di in SI .. Store.Last_List(Index) loop
        if j > S'last then return GT;end if;
        UD := Comparison_Map(Store.Data(di));
        US := Comparison_Map(S(j));
        if UD > US then return GT;
        elsif UD < US then return LT;
        end if;
        j := j+1;
      end loop;
      if j <= S'last then return LT;end if;
      return EQ;
    end if;
  end Compare;

  function To_Mapped(S : Data_Type;
                     Comparison_Map : Comparison_Map_Type) return String is
    Result : String(1 .. S'length);
  begin
    for i in S'range loop
      Result(Integer(i-S'first)+1) := Comparison_Map(S(i));
    end loop;
    return Result;
  end To_Mapped;

  function Compare(Store : Store_Type;
                   Comparison_Map : Comparison_Map_Type;
                   Left, Right : String_Indices)
  return Comparisons is
    use Ada.Characters.Handling;
    LI,RI : Data_Indices;
    LLen,RLen : Natural;
  begin
    if Left >= Store.Next_String or Right >= Store.Next_String then
      raise Nonesuch;
    end if;
    if Left = Right then return EQ;end if;
    LI := Store.Last_List(Left-1)+1;
    LLen := Natural(Store.Last_List(Left) - Store.Last_List(Left-1));
    RI := Store.Last_List(Right-1)+1;
    RLen := Natural(Store.Last_List(Right) - Store.Last_List(Right-1));
    if LLen = RLen
    and then To_Mapped(Store.Data(LI .. Store.Last_List(Left)), Comparison_Map)
             = To_Mapped(Store.Data(RI .. Store.Last_List(Right)), Comparison_Map) then
      return EQ;
    end if;
    if RLen = 0 then return GT;end if;
    if LLen = 0 then return LT;end if;
    -- not EQ and neither is zero length
    if Comparison_Map(Store.Data(LI)) > Comparison_Map(Store.Data(RI)) then return GT;
    elsif Comparison_Map(Store.Data(LI)) < Comparison_Map(Store.Data(RI)) then return LT;
    elsif To_Mapped(Store.Data(LI .. Store.Last_List(Left)), Comparison_Map)
          > To_Mapped(Store.Data(RI .. Store.Last_List(Right)), Comparison_Map) then
      return GT;
    elsif To_Mapped(Store.Data(LI .. Store.Last_List(Left)), Comparison_Map)
          < To_Mapped(Store.Data(RI .. Store.Last_List(Right)), Comparison_Map) then
      return LT;
    else return EQ;
    end if;
  end Compare;

  protected type Semaphores is
    entry Grab;
    procedure Free;
  private
    Available : Boolean := True;
  end Semaphores;

  protected body Semaphores is
    entry Grab when Available is
    begin
      Available := False;
    end Grab;
    procedure Free is
    begin
      Available := True;
    end Free;
  end Semaphores;

  package body Ordered_Store is

     type Ids_To_Indices_Type is array(IDs range <>) of String_Indices;

     Next  : Order_Positive := 1;
     Next_ID : IDs := IDs'first;
     IDs_to_Indices : IDs_To_Indices_Type(IDs);
     Order : Ordered_Lists(1 .. IDs_to_Indices'length);

    function Count return String_Counts is
    -- return the number of distinct strings in this instantiation.
    begin return String_Counts(Next-1);end Count;

    Semaphore : Semaphores;

    function Is_Present(S : in String) return Boolean is
      Min : Order_Positive := 1;
      Max : Order_Natural := Next-1;  -- zero for empty table
      Guess: Order_Positive;
    begin
      if Storage.Task_Safe then Semaphore.Grab;end if;
      while Min <= Max loop
        Guess := Min+(Max-Min)/2;
        case Compare(Storage, Comparison, IDs_to_Indices(Order(Guess)), S) is
          when LT => Min := Guess+1;
          when GT => Max := Guess-1;
          when EQ =>
            if Storage.Task_Safe then Semaphore.Free;end if;
            return True;
        end case;
      end loop;
      if Storage.Task_Safe then Semaphore.Free;end if;
      return False;
    exception
      when others =>
        if Storage.Task_Safe then Semaphore.Free;end if;
        raise;
    end Is_Present;

    function Put(S : String) return IDs is
    -- Add to store if it isn't there yet.  Return new or old ID.
    -- Raise Overflow if too many or too long strings.
      Min : Order_Positive := 1;
      Max : Order_Natural := Next-1;  -- zero for empty table
      Guess: Order_Positive;
      Index : String_Indices;
      Result : Ids;
    begin
      if Storage.Task_Safe then Semaphore.Grab;end if;
      while Min <= Max loop
        Guess := Min+(Max-Min)/2;
        case Compare(Storage, Comparison, IDs_to_Indices(Order(Guess)), S) is
          when LT => Min := Guess+1;
          when GT => Max := Guess-1;
          when EQ =>
            Result := Order(Guess);
            if Storage.Task_Safe then Semaphore.Free;end if;
            return Result;
        end case;
      end loop;
      if Next >= Order'last then
        raise Overflow;
      end if;
      Add(Storage, S, Index);
      Order(Min+1 .. Next) := Order(Min .. Next-1);
      Order(Min) := Next_ID;
      IDs_to_Indices(Next_ID) := Index;
      Next := Next+1;
      Next_ID := IDs'succ(Next_ID);
      Result := IDs'pred(Next_ID);
      if Storage.Task_Safe then Semaphore.Free;end if;
      return Result;
    exception
      when others =>
        if Storage.Task_Safe then Semaphore.Free;end if;
        raise;
    end Put;

    procedure Lookup(S : in String; Success : out Boolean; ID : out IDs) is
    -- if S is not Present, Success := False
    -- else Success := True and ID is set.
      Min : Order_Positive := 1;
      Max : Order_Natural := Next-1;  -- zero for empty table
      Guess: Order_Positive;
    begin
      if Storage.Task_Safe then Semaphore.Grab;end if;
      while Min <= Max loop
        Guess := Min+(Max-Min)/2;
        case Compare(Storage, Comparison, IDs_to_Indices(Order(Guess)), S) is
          when LT => Min := Guess+1;
          when GT => Max := Guess-1;
          when EQ =>
            if Storage.Task_Safe then Semaphore.Free;end if;
            Success := True;
            ID := Order(Guess);
            return;
        end case;
      end loop;
      if Storage.Task_Safe then Semaphore.Free;end if;
      Success := False;
    exception
      when others =>
        if Storage.Task_Safe then Semaphore.Free;end if;
        raise;
    end Lookup;

    function Get(ID : IDs) return String is
    -- raise Nonesuch if no such ID.
    begin
      if ID in Next_ID .. IDs'last then  -- we ain't there yet!
        raise Nonesuch;
      end if;
      return Get(Storage, IDs_to_Indices(ID));
    end Get;

    function Get_Length(ID : IDs) return Natural is
    -- raise Nonesuch if no such ID.
    begin
      if ID >=  Next_ID then
        raise Nonesuch;
      end if;
      return Get_Length(Storage, IDs_to_Indices(ID));
    end Get_Length;

    function Get_Ordered_Contents return Ordered_Lists is
    -- returns array of all IDs in table, in order.
    begin return Order(1 .. Next-1);end Get_Ordered_Contents;

    function Convert is new Ada.Unchecked_Conversion(Four_Bytes,
                                                     Interfaces.Integer_32);
    function Convert is new Ada.Unchecked_Conversion(Interfaces.Integer_32,
                                                     Four_Bytes);

    function External_Ids(Id : Ids) return Four_Bytes is
    begin
      return Convert(Interfaces.Integer_32(IDs'pos(ID)));
    end External_IDs;

    function Internal_Ids(X : Four_Bytes) return IDs is
    begin
      return IDs'val(Interfaces.Integer_32'(Convert(X)));
    end Internal_IDs;

    function External_Order_Natural(X : Order_Natural) return Four_Bytes is
    begin
      return Convert(Interfaces.Integer_32(X));
    end External_Order_Natural;

    function Internal_Order_Natural(X : Four_Bytes) return Order_Natural is
    begin
      return Order_Natural(Interfaces.Integer_32'(Convert(X)));
    end Internal_Order_Natural;

    procedure Ids_To_Indices_Writer(Stream : access Ada.Streams.Root_Stream_Type'class;
                                    Info   : in     Ids_To_Indices_Type) is
      use Ada.Streams;
      di : Stream_Element_Offset;
      Four : Four_Bytes;
      Buffer : Buffer_Type;
    begin
      di := 1;
      for i in Info'range loop
        if di > Buffer'last then
          Write(Stream.all, Buffer);
          di := 1;
        end if;
        Four := External_String_Counts(Info(i));
        Buffer(di .. di+3) := Four;
        di := di+4;
      end loop;
      if di > 1 then
        Write(Stream.all, Buffer(1 .. di-1));
      end if;
    end Ids_To_Indices_Writer;

    procedure Ids_To_Indices_Reader(Stream : access Ada.Streams.Root_Stream_Type'class;
                                    Info   : out    Ids_To_Indices_Type) is
      use Ada.Streams;
      Last,Last_To_Get : Stream_Element_Offset;
      si : Stream_Element_Offset;
      Four : Four_Bytes;
      Buffer : Buffer_Type;
    begin
      si := Buffer'last+1;
      for i in Info'range loop
        if si > Buffer'last then
          Last_To_Get := Stream_Element_Offset'min(Buffer'last,
                           4*Stream_Element_Offset
                             (IDs'pos(Info'last) - IDs'pos(i)+1));
          Read(Stream.all, Buffer(1 .. Last_To_Get), Last);
          if Last /= Last_To_Get then
            raise Ada.Streams.Stream_IO.Data_Error;
          end if;
          si := 1;
        end if;
        Four := Buffer(si .. si+3);
        Info(i) := Internal_String_Counts(Four);
        si := si+4;
      end loop;
    end Ids_To_Indices_Reader;

    procedure Order_Writer(Stream : access Ada.Streams.Root_Stream_Type'class;
                           Info   : in     Ordered_Lists) is
      use Ada.Streams;
      di : Stream_Element_Offset;
      Four : Four_Bytes;
      Buffer : Buffer_Type;
    begin
      di := 1;
      for i in Info'range loop
        if di > Buffer'last then
          Write(Stream.all, Buffer);
          di := 1;
        end if;
        Four := External_IDs(Info(i));
        Buffer(di .. di+3) := Four;
        di := di+4;
      end loop;
      if di > 1 then
        Write(Stream.all, Buffer(1 .. di-1));
      end if;
    end Order_Writer;

    procedure Order_Reader(Stream : access Ada.Streams.Root_Stream_Type'class;
                           Info   : out    Ordered_Lists) is
      use Ada.Streams;
      Last,Last_To_Get : Stream_Element_Offset;
      si : Stream_Element_Offset;
      Four : Four_Bytes;
      Buffer : Buffer_Type;
    begin
      si := Buffer'last+1;
      for i in Info'range loop
        if si > Buffer'last then
          Last_To_Get := Stream_Element_Offset'min(Buffer'last,
                           4*Stream_Element_Offset(Info'last-i+1));
          Read(Stream.all, Buffer(1 .. Last_To_Get), Last);
          if Last /= Last_To_Get then
            raise Ada.Streams.Stream_IO.Data_Error;
          end if;
          si := 1;
        end if;
        Four := Buffer(si .. si+3);
        Info(i) := Internal_IDs(Four);
        si := si+4;
      end loop;
    end Order_Reader;

    procedure Write(Stream : access Ada.Streams.Root_Stream_Type'class) is
    begin
      Ada.Streams.Write(Stream.all, External_Order_Natural(Next-1));
      Ada.Streams.Write(Stream.all, External_Ids(Next_ID));
      if Next_ID > IDs'first then
        IDs_To_Indices_Writer(Stream, IDs_To_Indices(IDs'first .. IDs'pred(Next_ID)));
      end if;
      if Next > 1 then
        Order_Writer(Stream, Order(1 .. Next-1));
      end if;
    end Write;

    procedure Read(Stream : access Ada.Streams.Root_Stream_Type'class) is
      use type Ada.Streams.Stream_Element_Offset;
      Four : Four_Bytes;
      New_Next : Order_Positive;
      Last : Ada.Streams.Stream_Element_Offset;
    begin
      Ada.Streams.Read(Stream.all, Four, Last);
      if Last /= Four'last then
        raise Ada.Streams.Stream_IO.Data_Error;
      end if;
      New_Next := Internal_Order_Natural(Four)+1;
      Ada.Streams.Read(Stream.all, Four, Last);
      if Last /= Four'last then
        raise Ada.Streams.Stream_IO.Data_Error;
      end if;
      Next_ID := Internal_Ids(Four);  -- Constraint Error if too big
      Next := New_Next;
      if Next_ID > IDs'first then
        IDs_To_Indices_Reader(Stream, IDs_To_Indices(IDs'first .. IDs'pred(Next_ID)));
      end if;
      if Next > 1 then
        Order_Reader(Stream, Order(1 .. Next-1));
      end if;
    end Read;

  end Ordered_Store;

end WORM_Str;
