with Claw,
     Claw.Sockets,
     Text_Line_Sockets;
package Te_Text_Line_Sockets is

  -- Transfer-Encoding handling Text_Line_Sockets
  -- currently handles only plain and chunked

  -- 07/20/01 fix zero size chunk handling per Randy

  type Encoding_Type is (Plain, Chunked);

  type Socket_Type     -- (Max_Line_Length: Positive) is inherited
  is new Text_Line_Sockets.Socket_Type with private;

  Bad_Chunk : Exception;  -- size of chunk not as specified

  Non_Text  : Exception renames Text_Line_Sockets.Non_Text;
   -- raised by Get_Line on apparent non-text data

  procedure Get_Line(Socket  : in out Socket_Type;
                     Line    :    out String;
                     Last    :    out Natural;
                     Timeout : in     Duration := 30.0);
    -- Return Line(Line'first .. Last)
    -- Input up through LF if LF occurs,
    -- Line(Line'range) if no LF seen,
    -- or Input up through last received char if timeout or hangup
    -- NOTE Timeout parameter is last (unlike Claw.Sockets.Get) for defaulting
    -- raises
    --   Bad_Chunk         if chunked and screwed up
    --   Non_Text          if data has excessive fraction of "control" characters

  procedure Set_Encoding(Socket : in out Socket_Type;
                         Encoding : in Encoding_Type);

  -- NOTE: Chunked coding must be set before reading any chunks of a set.
  -- It must be reset before reading another set of chunked encoding.
  -- Plain must be set at the beginning of plain coding.

  procedure Close(Socket : in out Socket_Type);

private
  type Socket_Type     -- (Max_Line_Length: Positive) is inherited
  is new Text_Line_Sockets.Socket_Type with record
    Encoding : Encoding_Type := Plain;
    Remaining_In_Chunk : Claw.Natural_Int;
    Last_Chunk_Seen : Boolean;  -- set True by zero size chunk
  end record;

end Te_Text_Line_Sockets;
