with Ada.Exceptions,
     Ada.Streams,
     Claw.Sockets;
package Text_Line_Sockets is

  type Socket_Type(Max_Line_Length: Positive)
  is new Claw.Sockets.Socket_Type with private;

  Non_Text : exception;

  procedure Get_Line(Socket  : in out Socket_Type;
                     Line    :    out String;
                     Last    :    out Natural;
                     Timeout : in     Duration := 30.0);
    -- Return Line(Line'first .. Last)
    -- Input up through LF if LF occurs,
    -- Line(Line'range) if no LF seen,
    -- or Input up through last received char if timeout or hangup
    -- NOTE Timeout parameter is last (unlike Claw.Sockets.Get) for defaulting
    -- Raises Non_Text if incoming data has excessive number of "control" characters

  procedure Close(Socket : in out Socket_Type);

--       following routines need full, not Intro version of Claw.Sockets
--procedure Input (Socket : in out Socket_Type;
--                 Item   :    out Ada.Streams.Stream_Element_Array;
--                 Last   :    out Ada.Streams.Stream_Element_Offset);
--procedure Input (Socket  : in out Socket_Type;
--                 Timeout : in     Duration;
--                 Item    :    out Ada.Streams.Stream_Element_Array;
--                 Last    :    out Ada.Streams.Stream_Element_Offset);
--procedure Input_All (Socket  : in out Socket_Type;
--                     Timeout : in     Duration := 120.0;
--                     Item    :    out Ada.Streams.Stream_Element_Array;
--                     Last    :    out Ada.Streams.Stream_Element_Offset);
--generic
--  type Element_Type is private;
--procedure Read (Socket : in out Socket_Type'class;
--                Item   :    out Element_Type);
--
--generic
--  type Element_Type is private;
--procedure Try_Read (Socket  : in out Socket_Type'class;
--                    Timeout : in     Duration;
--                    Item    :    out Element_Type);
--
--generic
--  type Element_Type is private;
--procedure Write (Socket : in out Socket_Type'class;
--                 Item   : in     Element_Type);

private

  type Statuses is (OK, Problem, EOF);

  type Socket_Type(Max_Line_Length: Positive)
  is new Claw.Sockets.Socket_Type with record
    Buffer : String(1 .. Max_Line_Length);
    Last   : Natural := 0;
    Next   : Positive := 1;
    Status : Statuses := OK;
    Non_Text_Fraction : Float := 0.0;
    Occurrence : Ada.Exceptions.Exception_Occurrence;
  end record;

  procedure Finalize(Socket : in out Socket_Type);

end Text_Line_Sockets;
